(* library function's type envrionment *)
open Alpha

type envs =
    {for_typing : string * Typing.typ;
     for_alpha : string * ident;
     for_typeinf : ident * typ;
     for_m2s_typeinf : ident * typ}

(* id  *)
let new_negative_int =
  let counter = ref (-1) in
    fun () -> decr counter; !counter

let rec typ_alpha_to_typing typ =
  match typ with
      TString -> Typing.TString
    | TInt -> Typing.TInt
    | TBool -> Typing.TBool
    | TLabel -> Typing.TLabel
    | TAttribute -> Typing.TAttribute
    | TFun(typ1, typ2) -> Typing.TFun((typ_alpha_to_typing typ1), (typ_alpha_to_typing typ2))
    | TTree u -> Typing.TTree
    | TTreelist u -> Typing.TTreelist

let use_plus_to_omega use =
  match use with
      UPlus -> UOmega
    | UOmega | UOne -> use

let rec typ_plus_to_omega typ =
  match typ with
      TString -> TString
    | TInt -> TInt
    | TBool -> TBool
    | TLabel -> TLabel
    | TAttribute -> TAttribute
    | TFun(typ1, typ2) -> TFun((typ_plus_to_omega typ1), (typ_plus_to_omega typ2))
    | TTree u -> TTree (use_plus_to_omega u)
    | TTreelist u -> TTreelist (use_plus_to_omega u)

let make_elem (value_name : string) (typ : typ) = 
  let id_num = new_negative_int () in
  let alpha_id = {id = id_num; origin = value_name} in
  let for_typing = (value_name, (typ_alpha_to_typing typ)) in
  let for_alpha = (value_name, alpha_id) in
  let for_typeinf = (alpha_id, (typ_plus_to_omega typ)) in
  let for_m2s_typeinf = (alpha_id, typ) in
    {for_typing = for_typing;
     for_alpha = for_alpha;
     for_typeinf = for_typeinf;
     for_m2s_typeinf = for_m2s_typeinf}
  
let env_list =
  (make_elem "parse_list" (TFun((TTreelist UOne), (TTreelist UOmega))))::
    (make_elem "print_list" (TFun((TTreelist UOmega), TInt)))::
    (make_elem "sort" (TFun( TFun(TTree UOmega, TFun(TTree UOmega, TInt)),
			     TFun(TTreelist UOmega, TTreelist UOmega))))::
    (make_elem "lessthan"  (TFun((TTree UOmega),  TFun((TTree UOmega), TInt))))::
    (make_elem "int_of_string" (TFun(TString, TInt)))::
    (make_elem "string_of_int" (TFun(TInt, TString)))::
    (make_elem "pc_of_int" (TFun(TInt, (TTree UPlus))))::
    (make_elem "append_outlist"
      (TFun(TTreelist UPlus,
	    TFun(TTreelist UPlus, TTreelist UPlus))))::
    []

let init_env_typing = 
  List.map (fun t -> t.for_typing) env_list
let init_map_alpha =
  List.map (fun t -> t.for_alpha) env_list
let init_env_typeinf =
  List.map (fun t -> t.for_typeinf) env_list
let init_env_m2s_typeinf =
  List.map (fun t -> t.for_m2s_typeinf) env_list
