open Typeinf


(* debug *)
let rec print_ans ans =
  match ans with
      [] -> print_string "[]"
    | (uid, u)::tl ->
	print_string "(";
	print_int uid;
	print_string "=";
	print_usage u;
	print_string ")::";
	print_ans tl

let subst_use u_source u_target c =
  match c with
      Ge_use(u1, u2) ->
	let u1' = if u1 = Alpha.UVar u_target then u_source else u1 in
	let u2' = if u2 = Alpha.UVar u_target then u_source else u2 in
	  Ge_use(u1', u2')
    | Eq_use(u1, u2) ->
	let u1' = if u1 = Alpha.UVar u_target then u_source else u1 in
	let u2' = if u2 = Alpha.UVar u_target then u_source else u2 in
	  Eq_use(u1', u2')

let find_use u ans =
  try
    List.assoc u ans
  with
    | _ -> Alpha.UPlus


let rec transitive constrs use1 =
  match constrs with
      (Ge_use(u1, u2))::tl when (u1 = use1) ->
	(Ge_use(Alpha.UOmega, u2))::(transitive tl use1)
    | hd::tl ->
	transitive tl use1
    | [] -> []
     
let rec below_omega tobe_cs done_cs ans changed =
(*
print_string "omega\n";
Typeinf.print_constr tobe_cs;
print_newline();
*)
  match tobe_cs with
      [] ->
	if changed then
	  below_omega done_cs [] ans false
	else ans
    | Ge_use(Alpha.UVar uv1, Alpha.UOmega)::tl ->
	below_omega tl done_cs ans changed
    | Ge_use(Alpha.UOmega, Alpha.UVar uv2)::tl ->
	let tobe_cs' = List.map (subst_use Alpha.UOmega uv2) tl in
	let done_cs' = List.map (subst_use Alpha.UOmega uv2) done_cs in
	  below_omega tobe_cs' done_cs' ((uv2, Alpha.UOmega)::ans) true
    | hd::tl ->
	below_omega tl (hd::done_cs) ans changed

let top_below_omega constrs = 
  below_omega constrs [] [] false

let rec subst_ans ans substed_list =
  match substed_list with
      (uvar1, uvar2)::tl ->
	if (List.mem uvar1 (List.map fst ans)) then
	  subst_ans ((uvar2, (List.assoc uvar1 ans))::ans) tl
	else subst_ans ans tl
    | [] -> ans

let rec solver_iter tobe_cs done_cs ans substed substed_list =
  match tobe_cs with
      [] -> 
	(
	if substed then
	    solver_iter done_cs [] ans false substed_list
	else	  
	  begin
	    let ans' = top_below_omega done_cs in
	    let ans = ans@ans' in
	      subst_ans ans substed_list
	  end
	)
    | (Ge_use(u1, u2))::tl ->
	  begin
	    match u1, u2 with
		Alpha.UPlus, _ | _, Alpha.UOne ->
		  solver_iter tl done_cs ans substed substed_list
	      | u1, u2 when u1=u2 ->
		  solver_iter tl done_cs ans substed substed_list
	      | Alpha.UOmega, Alpha.UVar u ->
		  solver_iter tl ((Ge_use(Alpha.UOmega, Alpha.UVar u))::done_cs) ans substed substed_list
	      | Alpha.UOne, Alpha.UVar u ->
		  let tl' = List.map (subst_use Alpha.UOne u) tl in
		  let done_cs' = List.map (subst_use Alpha.UOne u) done_cs in
		    solver_iter tl' done_cs' ((u, Alpha.UOne)::ans) true substed_list
	      | Alpha.UVar u, Alpha.UPlus ->
		  let tl' = List.map (subst_use Alpha.UPlus u) tl in
		  let done_cs' = List.map (subst_use Alpha.UPlus u) done_cs in
		    solver_iter tl' done_cs' ((u, Alpha.UPlus)::ans) true substed_list
	      | Alpha.UVar u, u2 ->
		  solver_iter tl ((Ge_use(Alpha.UVar u, u2))::done_cs) ans substed substed_list
	      | _ -> failwith "can't solve constr"
	  end 
    | (Eq_use(u1, u2))::tl ->
	begin
	  match u1, u2 with
	      use1, use2 when use1 = use2 -> solver_iter tl done_cs ans substed substed_list
	    | Alpha.UVar uvar1, Alpha.UVar uvar2 -> 
		let tl' = List.map (subst_use (Alpha.UVar uvar1) uvar2) tl in
		let done_cs' = List.map (subst_use (Alpha.UVar uvar1) uvar2) done_cs in
		  solver_iter tl' done_cs' ans true ((uvar1, uvar2)::substed_list)
	    | Alpha.UVar uvar1, u2 ->
		let tl' = List.map (subst_use u2 uvar1) tl in
		let done_cs' = List.map (subst_use u2 uvar1) done_cs in
		  solver_iter tl' done_cs' ((uvar1, u2)::ans) true substed_list
	    | u1, Alpha.UVar uvar2 ->
		let tl' = List.map (subst_use u1 uvar2) tl in
		let done_cs' = List.map (subst_use u1 uvar2) done_cs in
		  solver_iter tl' done_cs' ((uvar2, u1)::ans) true substed_list
	    | _ -> failwith "can't solve constr"
	end
(*
let rec solver_iter tobe_cs done_cs ans substed =
  match tobe_cs with
      [] ->
	(if substed then
	  solver_iter done_cs [] ans false
	else
	  ans)
    | (Ge_use(Alpha.UPlus, _))::tl
    | (Ge_use(Alpha.UOmega, Alpha.UOne))::tl ->
	solver_iter tl done_cs ans substed
    | (Ge_use(u1, u2))::tl when u1=u2 ->
	solver_iter tl done_cs ans substed
    | (Ge_use(Alpha.UOmega, Alpha.UVar u))::tl ->
	solver_iter tl ((Ge_use(Alpha.UOmega, Alpha.UVar u))::done_cs) ans substed
    | (Ge_use(Alpha.UOne, Alpha.UVar u))::tl ->
	let tl' = List.map (subst_use Alpha.UOne u) tl in
	let done_cs' = List.map (subst_use Alpha.UOne u) done_cs in
	  solver_iter tl' done_cs' ((u, Alpha.UOne)::ans) true
    | (Ge_use(Alpha.UVar u, Alpha.UPlus))::tl ->
	let tl' = List.map (subst_use Alpha.UPlus u) tl in
	let done_cs' = List.map (subst_use Alpha.UPlus u) done_cs in
	  solver_iter tl' done_cs' ((u, Alpha.UPlus)::ans) true
    | (Ge_use(Alpha.UVar u, u2))::tl ->
	solver_iter tl ((Ge_use(Alpha.UVar u, u2))::done_cs) ans substed
    | _ -> failwith "can't solve constr"
*)




let rec remake_constr constr =
  match constr with
      (Ge_use(use1, use2))::tl ->
	(match use1, use2 with
	     Alpha.UVar u1, u2 -> (Ge_use(Alpha.UVar u1, u2))::(remake_constr tl)
	   | u1, Alpha.UVar u2 -> (Ge_use(u1, Alpha.UVar u2))::(remake_constr tl)
	   | Alpha.UPlus, _ 
	   | Alpha.UOmega, Alpha.UOmega | Alpha.UOmega, Alpha.UOne
	   | Alpha.UOne, Alpha.UOne -> remake_constr tl
	   | _ -> failwith "can't solve constr")
    | [] -> []



let rec solver (constrs : Typeinf.constr list)=
    solver_iter constrs [] [] false []

let rec translate t constrs =
  let ans = solver constrs in
    (trans_inner ans t)

and trans_inner ans (t, typ) =
  let pre_t =
    match t with
	String s -> t
      | Int n -> t
      | True | False -> t
      (* | Pcdata pc -> t *)
      | Label l -> t
      | Attribute a -> t
      | AttrGet (a, x) -> 
	  AttrGet((trans_inner ans a), x)
      | True -> t
      | False -> t
      | BinOp (binop, t1, t2) ->
	  BinOp(binop, trans_inner ans t1, trans_inner ans t2)
      | Var x -> t
      | Fix(f, x, t') ->
	  Fix(f, x, trans_inner ans t')
      | App(t1, t2) ->
	  App(trans_inner ans t1, trans_inner ans t2)
      | Let(x, t1, t2) ->
	  Let(x, (trans_inner ans t1), (trans_inner ans t2))
      | Letrecs(fts, t) ->
	  let rec trans_inner_match ans fts =
	    (match fts with
		 (fn, tn)::tl ->
		   (fn, (trans_inner ans tn))::(trans_inner_match ans tl)
	       | [] -> [])
	  in 
	  let fts' = trans_inner_match ans fts in
	  let t' = trans_inner ans t in
	    Letrecs(fts', t')
      | If(t0, t1, t2) ->
	  If(trans_inner ans t0, trans_inner ans t1, trans_inner ans t2)
      | Leaf(u, t) ->
	  Leaf(translate_use u ans, trans_inner ans t)
      | Node(u, t1, t2, t3) ->
	  Node(translate_use u ans, trans_inner ans t1, trans_inner ans t2, trans_inner ans t3)
      | Nil(u) -> Nil(translate_use u ans)
      | Cons(u, t1, t2) -> 
	  Cons(translate_use u ans, trans_inner ans t1, trans_inner ans t2)
(**)
      | Match(typ, id, match_list) -> 
	  let rec trans_inner_match ans (p, t) =
	    (p, (trans_inner ans t))
	  in
	  (
	  match typ with 
	      Alpha.TTree u ->
		Match(Alpha.TTree (translate_use u ans), id, (List.map (trans_inner_match ans) match_list))
	    | Alpha.TTreelist u ->
		Match(Alpha.TTreelist (translate_use u ans), id, (List.map (trans_inner_match ans) match_list))
	    | _ -> Match(typ, id, (List.map (trans_inner_match ans) match_list))
	  ) 
      | Coerce(env1, env2, t) -> failwith "M2S_SOLVER.trans_inner : Coerce cannot occur"
      | CoerceOut(typ1, typ2, t) ->
	  (match (typ1, typ2) with
	       (Alpha.TTree u1), (Alpha.TTree u2) ->
		 let typ1' = (Alpha.TTree (translate_use u1 ans)) in
		 let typ2' = (Alpha.TTree (translate_use u2 ans)) in
		   translate_coerce typ1' typ2' (trans_inner ans t)
	     | (Alpha.TTreelist u1), (Alpha.TTreelist u2) ->
		 let typ1' = (Alpha.TTreelist (translate_use u1 ans)) in
		 let typ2' = (Alpha.TTreelist (translate_use u2 ans)) in
		   translate_coerce typ1' typ2' (trans_inner ans t)
	     | _ -> (fst (trans_inner ans t)))
      | LetS2M_t(x, t) ->
	  LetS2M_t(x, (trans_inner ans t))
      | LetS2M_tl(x, tl) ->
	  LetS2M_tl(x, (trans_inner ans tl))
  in
    pre_t, (translate_type typ ans) 
(*
  | M2S _ -> failwith "translate: M2S is not allowed to occur here"
  | Copy _ -> failwith "translate: Copy is not allowed to occur here"
*)
and translate_type typ ans =
  match typ with
      Alpha.TString | Alpha.TInt | Alpha.TBool | Alpha.TLabel | Alpha.TAttribute -> typ
    | Alpha.TFun(t1, t2) ->
	Alpha.TFun(translate_type t1 ans, translate_type t2 ans)
    | Alpha.TTree u -> Alpha.TTree (translate_use u ans)
    | Alpha.TTreelist u -> Alpha.TTreelist (translate_use u ans)
and translate_use u ans =
  match u with
      Alpha.UOne | Alpha.UOmega | Alpha.UPlus -> u
    | Alpha.UVar u -> find_use u ans
and translate_coerce typ1 typ2 t =
  let pre_t, typ = t in
    match typ1, typ2 with
	Alpha.TString,  Alpha.TString
      | Alpha.TInt, Alpha.TInt
      | Alpha.TBool, Alpha.TBool
      | Alpha.TLabel, Alpha.TLabel | Alpha.TAttribute, Alpha.TAttribute -> pre_t
      | Alpha.TFun(t1d, t1cod), Alpha.TFun(t2d, t2cod) when t1d = t2d && t1cod = t2cod -> pre_t
      | Alpha.TTree(u1), Alpha.TTree(u2) when u1 = u2 -> pre_t
      | Alpha.TTree(Alpha.UOmega), Alpha.TTree(Alpha.UPlus) ->
	  M2S_t(t)
      | Alpha.TTreelist(u1), Alpha.TTreelist(u2) when u1 = u2 -> pre_t
      | Alpha.TTreelist(Alpha.UOmega), Alpha.TTreelist(Alpha.UPlus) ->
	  M2S_tl(t)
(*
      | _ -> 
	  print_typ typ1;print_newline();print_typ typ2;print_newline();
	  print_term 0 t;
	  print_newline();
	  failwith "M2S_SOLVER.translate_coerce: CANNOT OCCUR!"
*)
