open Typeinf

let rec rename_typ typ = 
  match typ with
      Alpha.TTree _ -> Alpha.TTree (Alpha.new_uvar ())
    | Alpha.TTreelist _ -> Alpha.TTreelist (Alpha.new_uvar ())
    | Alpha.TFun(typ1, typ2) ->	Alpha.TFun(typ1, rename_typ typ2)
    | _ -> typ

let rec above_than_below_typ typ1 typ2 =
  let u1 = use_of typ1 in
  let u2 = use_of typ2 in
    if u1 = u2 then
      []
    else
      [(Ge_use(u1, u2))]

let rec unify_typ typ1 typ2 =
  match (typ1, typ2) with
      typ1, typ2 when typ1=typ2 -> []
    | Alpha.TFun(typ1d, typ1cod), Alpha.TFun(typ2d, typ2cod) ->
	(unify_typ typ1d typ2d) @ (unify_typ typ1cod typ2cod)
    | Alpha.TTree u1, Alpha.TTree u2 ->
	unify_use u1 u2
    | Alpha.TTreelist u1, Alpha.TTreelist u2 ->
	unify_use u1 u2
    | _ -> print_typ typ1; print_newline();  print_typ typ2; failwith "m2s_unify"
and unify_use use1 use2 =
  match (use1, use2) with
      u1, u2 when u1 = u2 -> []
    | _ -> [(Eq_use(use1, use2))]

let rec translate (env:env) (term:Typeinf.t) (typ:Alpha.typ) =
  let term', typ_of_term = term in  
  let t, constr =
    match term' with
	Var x ->
	  let c = unify_typ (List.assoc x env) typ in
	    Var x, c
      | String s -> String s, []
      | Int n -> Int n, []
      | True -> True, []
      | False -> False, []
      (* | Pcdata pc -> Pcdata pc, [] *)
      | Label l -> Label l, []
      | Attribute a -> Attribute a, []
      | AttrGet(a, x) ->
	  let a', c = translate env a Alpha.TAttribute in
	    AttrGet(a', x), c
	  (* binop = Eq | Sub | Cat | ListCat | Semi | Mod *)
      | BinOp(Eq, t1, t2) ->
	  let t1', c1 = translate env t1 (snd t1) in
	  let t2', c2 = translate env t2 (snd t2) in
	    BinOp(Eq, t1', t2'), (c1@c2)
      | BinOp(Semi, t1, t2) ->
	  let t1', c1 = translate env t1 (snd t1) in
	  let t2', c2 = translate env t2 typ in
	    BinOp(Semi, t1', t2'), (c1@c2)
      | BinOp(binop, t1, t2) ->
	  let typ_t1' = rename_typ (snd t1) in
	  let typ_t2' = rename_typ (snd t2) in
	  let t1', c1 = translate env t1 typ_t1' in
	  let t2', c2 = translate env t2 typ_t2' in
	  let c3 = above_than_below_typ typ typ_t1' in
	  let c4 = above_than_below_typ typ typ_t2' in
	  let t1'' = CoerceOut(typ_t1', typ, t1') in  
	  let t2'' = CoerceOut(typ_t2', typ, t2') in  
	    BinOp(binop, (t1'', typ), (t2'', typ)), (c1@c2@c3@c4)
      | App(t1, t2) ->
	  let typ2 = rename_typ (snd t2) in
	  let u2 = use_of typ2 in
	  let c0 = [](*[Ge_use(Alpha.UOmega, u2)]*) in
	  let t1', c1 = translate env t1 (Alpha.TFun(typ2, typ)) in
	  let t2', c2 = translate env t2 typ2 in
	    App(t1', t2'), (c0@c1@c2)
(*
      | Fix(f, x, body) ->
	  (match typ with
	       (Alpha.TFun(typ_d, typ_cod)) as typ_f ->
		 let env' = ((f, typ_f)::(x, typ_d)::env) in
		 let typ_cod' = rename_typ typ_cod in
		 let c0 = above_than_below_typ typ_cod typ_cod' in
		 let body', c1 = translate env' body typ_cod' in
		 let c2 = [Ge_use((use_of typ_cod'), Alpha.UOmega)] in
		 let c3 = [Ge_use(Alpha.UOmega, (use_of typ_d))] in
		 let t' = Fix(f, x, (CoerceOut(typ_cod', typ_cod, body'), typ_cod)) in
		   t', (c0@c1@c2@c3))
*)
      | Fix(f, (x, u), body) ->
	  (match typ with
	       (Alpha.TFun(typ_d, typ_cod)) as typ_f ->
		 let env' = ((f, typ_f)::(x, typ_d)::env) in
		 (*let typ_cod' = rename_typ typ_cod in *)
		 let typ_cod' = typ_cod in
		 let c0 = above_than_below_typ typ_cod typ_cod' in
		 let body', c1 = translate env' body typ_cod' in
		 let c2 = [Ge_use((use_of typ_cod'), Alpha.UOmega)] in
		 let c3 = [Ge_use(Alpha.UOmega, (use_of typ_d))] in
		 let t' = Fix(f, (x, u), (CoerceOut(typ_cod', typ_cod, body'), typ_cod)) in
		   t', (c0@c1@c2@c3))
      | Let((x,u), t1, t2) ->
	  let typ_t1 = snd t1 in
	  let typ_t1' = rename_typ typ_t1 in
	  let t1', c0 = translate env t1 typ_t1' in
	  let c3 = [Eq_use(use_of typ_t1', Alpha.UOmega)] in
	  let env' = (x, typ_t1')::env in
	  let typ' = rename_typ typ in
	  let c1 = above_than_below_typ typ typ' in
	  let t2', c2 = translate env' t2 typ' in
	  let t2'' = CoerceOut(typ', typ, t2') in
	  Let((x,u), t1', (t2'', typ)), (c0@c1@c2@c3)
      | Letrecs (fts, t) ->
	  let rec rename_fts fts =
	    match fts with
		(fn, (tn, typ_tn))::tl ->
		  (fn, (tn, rename_typ typ_tn))::(rename_fts tl)
	      | [] -> []
	  in
	  let fts' = rename_fts fts in
	  let rec env_letrecs fts env =
	    (match fts with
		 (fn, tn)::tl ->
		   (fn, (snd tn))::(env_letrecs tl env)
	       | [] -> env)
	  in
	  let env' = env_letrecs fts' env in
	  let rec trans_letrecs fts =
	    (match fts with
		 (fn, tn)::tl ->
		   let Alpha.TFun(typ_dom, typ_cod) as typ_tn = List.assoc fn env' in
		   let cn0 = [Ge_use(Alpha.UOmega, (use_of typ_dom))] in
		   let cn1 = [Ge_use((use_of typ_cod), Alpha.UOmega)] in
		   let tn', cn = translate env' tn typ_tn in
		     ((fn, tn'), cn@cn0@cn1)::(trans_letrecs tl)
		 | [] -> [])
	  in
	  let fts_cs = trans_letrecs fts' in
	  let fts' = List.map fst fts_cs in
	  let cs = List.map snd fts_cs in
	  let c = match cs with
	      hd::tl -> List.fold_left (@) hd tl
	  in
	  let t', c' = translate env' t typ in
	  let c = c@c' in
	    Letrecs(fts', t'), c
(*
      | Letrecs (fts, t) ->
	  let rec rename_fts fts =
	    match fts with
		(fn, (tn, typ_tn))::tl ->
		  (fn, (tn, rename_typ typ_tn))::(rename_fts tl)
	      | [] -> []
	  in
	  let fts' = rename_fts fts in
	  let rec env_letrecs fts env =
	    (match fts with
		 (fn, tn)::tl ->
		   (fn, (snd tn))::(env_letrecs tl env)
	       | [] -> env)
	  in
	  let env' = env_letrecs fts' env in
	  let rec trans_letrecs fts =
	    (match fts with
		 (fn, tn)::tl ->
		   let Alpha.TFun(typ_dom, typ_cod) as typ_tn = List.assoc fn env' in
		   let cn0 = [Ge_use(Alpha.UOmega, (use_of typ_dom)), AboveBelow] in
		   let cn1 = [Ge_use((use_of typ_cod), Alpha.UOmega), AboveBelow] in
		   let tn', cn = translate env' tn typ_tn in
		     ((fn, tn'), cn0@cn1@cn)::(trans_letrecs tl)
		 | [] -> [])
	  in
	  let fts_cs = trans_letrecs fts' in
	  let fts' = List.map fst fts_cs in
	  let cs = List.map snd fts_cs in
	  let c = match cs with
	      hd::tl -> List.fold_left (@) hd tl
	  in
	  let t', c' = translate env' t typ in
	  let c = c@c' in
	    Letrecs(fts', t'), c
*)

      | Leaf (Alpha.UOmega, pc) ->
	  let pc', c0 = translate env pc (Alpha.TString) in
	  let use = use_of typ in
	    Leaf(use, pc'), c0
      | Node(Alpha.UOmega, t1, t2, t3) ->
	  let use = use_of typ in
	  let typ3' = rename_typ (snd t3) in
	  let t1', c1 = translate env t1 (Alpha.TLabel) in
	  let t2', c2 = translate env t2 (Alpha.TAttribute) in
	  let t3', c3 = translate env t3 typ3' in
	  let c4 = above_than_below_typ (Alpha.TTreelist use) typ3' in
	  let t3'' = CoerceOut(typ3', (Alpha.TTreelist use), t3') in
	    Node(use, t1', t2', (t3'', Alpha.TTreelist use)), (c1@c2@c3@c4)
      | Cons(Alpha.UOmega, t1, t2) ->
	  let use = use_of typ in
	  let typ1' = rename_typ (snd t1) in
	  let c1 = above_than_below_typ (Alpha.TTree use) typ1' in
	  let typ2' = rename_typ (snd t2) in
	  let c2 = above_than_below_typ (Alpha.TTreelist use) typ2' in
	  let t1', c3 = translate env t1 typ1' in
	  let t2', c4 = translate env t2 typ2' in
	  let t1'' = CoerceOut(typ1', (Alpha.TTree use), t1') in
	  let t2'' = CoerceOut(typ2', (Alpha.TTreelist use), t2') in
	    (Cons(use, (t1'', (Alpha.TTree use)), (t2'', (Alpha.TTreelist use)))), (c1@c2@c3@c4) 
      | Nil (Alpha.UOmega) -> 
	  let use = use_of typ in
          Nil(use), []
(*
      | Match (typ_match, y, matchs) ->
	  let use_match = use_of typ_match in
	  let rec translate_match env (p, t) =
	    match p with
		Alpha.LeafPattern pc ->
		  let env' =
		    (Alpha.make_pattern_env pc Alpha.TString)@env in
		  let typ' = rename_typ typ in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env' t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1
	      | Alpha.NodePattern (p1, p2, p3) ->
		  let env' =
		    (Alpha.make_pattern_env p1 Alpha.TLabel)
		    @(Alpha.make_pattern_env p2 Alpha.TAttribute)
		    @(Alpha.make_pattern_env p3 (Alpha.TTreelist use_match))
		    @env
		  in
		  let typ' = rename_typ typ in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env' t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1
	      | Alpha.NilPattern
	      | Alpha.LabelPattern _ | Alpha.StringPattern _ ->
		  let typ' = rename_typ typ in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c
	      | Alpha.ConsPattern (p1, p2) ->
		  let env' =
		    (Alpha.make_pattern_env p1 (Alpha.TTree use_match))
		    @(Alpha.make_pattern_env p2 (Alpha.TTreelist use_match))
		    @env
		  in
		  let typ' = rename_typ typ in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env' t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1
	      | Alpha.IdentPattern id ->
		  let env' = (id, typ_match)::env in
		  let typ' = rename_typ typ in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env' t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1
	  in
	  let a = List.map (translate_match env) matchs in
	  let p_ts = List.map fst a in
	  let cs = List.map snd a in
	  let c = match cs with
	      hd::tl -> List.fold_left (@) hd tl
	  in
	    Match(typ_match, y, p_ts), c
*)
      | Match (typ_match, y, matchs) ->
	  let use_match = use_of typ_match in
	  let rec translate_match env (p, t) =
	    match p with
		Alpha.LeafPattern pc ->
		  let env' =
		    (Alpha.make_pattern_env pc Alpha.TString)@env in
		  let typ' = rename_typ (snd t) in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env' t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1
	      | Alpha.NodePattern (p1, p2, p3) ->
		  let env' =
		    (Alpha.make_pattern_env p1 Alpha.TLabel)
		    @(Alpha.make_pattern_env p2 Alpha.TAttribute)
		    @(Alpha.make_pattern_env p3 (Alpha.TTreelist use_match))
		    @env
		  in
		  let typ' = rename_typ (snd t) in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env' t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1
	      | Alpha.NilPattern
	      | Alpha.LabelPattern _ | Alpha.StringPattern _ ->
		  let typ' = rename_typ typ in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1
	      | Alpha.ConsPattern (p1, p2) ->
		  let env' =
		    (Alpha.make_pattern_env p1 (Alpha.TTree use_match))
		    @(Alpha.make_pattern_env p2 (Alpha.TTreelist use_match))
		    @env
		  in
		  let typ' = rename_typ (snd t) in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env' t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1
	      | Alpha.IdentPattern id ->
		  let env' = (id, typ_match)::env in
		  let typ' = rename_typ (snd t) in
		  let c1 = above_than_below_typ typ typ' in
		  let t', c = translate env' t typ' in
		    (p, (CoerceOut(typ', typ, t'), typ)), c@c1


	  in
	  let a = List.map (translate_match env) matchs in
	  let p_ts = List.map fst a in
	  let cs = List.map snd a in
	  let c = match cs with
	      hd::tl -> List.fold_left (@) hd tl
	  in
	    Match(typ_match, y, p_ts), c
      | If(t0, t1, t2) ->
	  let t0', c0 = translate env t0 (Alpha.TBool) in
	  let typ1' = rename_typ typ in
	  let typ2' = rename_typ typ in
	  let t1', c1 = translate env t1 typ1' in
	  let t2', c2 = translate env t2 typ2' in
	  let c3 = above_than_below_typ typ typ1' in
	  let c4 = above_than_below_typ typ typ2' in
	  let t1'' = CoerceOut(typ1', typ, t1') in
	  let t2'' = CoerceOut(typ2', typ, t2') in
	    If(t0', (t1'', typ), (t2'', typ)), (c0@c1@c2@c3@c4)
      | LetS2M_t(x, t) ->
	  let env' = (x, (Alpha.TTree Alpha.UOmega))::env in
	  let typ' = rename_typ typ in
	  let c0 = above_than_below_typ typ typ' in
	  let t', c1 = translate env' t typ' in
	    CoerceOut(typ', typ, (LetS2M_t(x, t'), typ)), (c0@c1)
      | LetS2M_tl(x, t) ->
	  let env' = (x, (Alpha.TTreelist Alpha.UOmega))::env in
	  let typ' = rename_typ typ in
	  let c0 = above_than_below_typ typ typ' in
	  let t', c1 = translate env' t typ' in
	    CoerceOut(typ', typ, (LetS2M_tl(x, t'), typ)), (c0@c1)
      | _ -> failwith "m2s_typeinf"  
  in
    (t, typ), constr
