open Typing

let debug_out s = prerr_string s; flush stderr

let init_env_typing = Libraryenv.init_env_typing
let init_map_alpha = Libraryenv.init_map_alpha
let init_env_typeinf = Libraryenv.init_env_typeinf
let init_env_m2s_typeinf = Libraryenv.init_env_m2s_typeinf

exception TypeMismatch of Alpha.typ

let main filename =
  let in_channel = open_in filename in
  let parsed = Parser.file Lexer.token (Lexing.from_channel in_channel) in
  let typed = Typing.infer init_env_typing parsed in
  let alpha = Alpha.alpha init_map_alpha typed in

  let new_ttl typ =
    match typ with
      	Alpha.TTree _ -> Alpha.TTree (Alpha.UOne)
      | Alpha.TTreelist _ -> Alpha.TTreelist (Alpha.UOne)
      | Alpha.TString -> Alpha.TString
      | Alpha.TInt -> Alpha.TInt
  in

  let omega_ttl typ =
    match typ with
	Alpha.TTree _ -> Alpha.TTree (Alpha.UOmega)
      | Alpha.TTreelist _ -> Alpha.TTreelist (Alpha.UOmega)
      | Alpha.TString -> Alpha.TString
      | Alpha.TInt -> Alpha.TInt
  in    
  let rec make_main_typ typ =
    match typ with 
	Alpha.TFun (typ0, Alpha.TFun(typ1, typ2)) ->
	  Alpha.TFun (new_ttl typ0, (make_main_typ (Alpha.TFun(typ1, typ2))))
      | Alpha.TFun (typ0, typ1) ->
	  Alpha.TFun (new_ttl typ0, omega_ttl typ1)
      | _ -> raise (TypeMismatch(snd alpha))
  in

  let typ = make_main_typ (snd alpha) in
  let translated, constr = 
    Typeinf.translate init_env_typeinf alpha typ
  in
  let a = Solver.translate translated constr in

  let plus_ttl typ =
    match typ with
      	Alpha.TTree _ -> Alpha.TTree (Alpha.UPlus)
      | Alpha.TTreelist _ -> Alpha.TTreelist (Alpha.UPlus)
      | Alpha.TString -> Alpha.TString
      | Alpha.TInt -> Alpha.TInt
  in
  let rec make_main_typ2 typ =
    match typ with 
	Alpha.TFun (typ0, Alpha.TFun(typ1, typ2)) ->
	  Alpha.TFun (typ0, (make_main_typ2 (Alpha.TFun(typ1, typ2))))
      | Alpha.TFun (typ0, typ1) ->
	  Alpha.TFun (typ0, plus_ttl typ1)
      | _ -> raise (TypeMismatch(snd alpha))
  in
  let typ = make_main_typ2 (snd a) in
  let b, constr = M2s_typeinf.translate init_env_m2s_typeinf a typ in

  let c = M2s_solver.translate b constr in
  let d = Trans_m2s.trans_m2s c in  
  let e = Trans_m2s.trans_copy d in
  let f = Trans_m2s.trans_skip e in
  (* Typeinf.print_term 0 f *)

  let constrs0 = Constr_gen.gen_constr f in
  let constrs1 = Constr_solv.solv_constr1 constrs0 in
  let constrs2 = Constr_solv.solv_constr2 constrs1 in
  let constrs3 = Constr_solv.reduce_constr constrs2 in
  let g = Part_buf.insert_buf f constrs3 in

  let h = Translation.translate (fst g) in
  (* let h = Translation.App(h, Translation.Unit) in *)
  let _ = GenerateCode.generate Format.std_formatter h in
  ()

let _ = 
  if !Sys.interactive then
    ()
  else
    begin
      main Sys.argv.(1);
      print_newline ()
    end
;;
