%{
(* Ūѿؿʤɤ *)
open Syntax

let rec make_cons = function
    [] -> Nil
  | hd::tl -> Cons(hd, make_cons tl)

let rec make_lam ids t =
  match ids with
      [hd] -> Lam(hd, t)
    | hd::tl -> Lam(hd, make_lam tl t)

%}

/* ȡ */
%token EOF
%token <string> IDENT
%token <string> STRING
%token <int> INT
%token LPAREN
%token RPAREN
%token COMMA
%token BEGIN
%token END
%token FUN
%token REC
%token LEAF
%token NODE
%token ARROW
%token LET
%token FIX
%token EQUAL
%token IN
%token MLEAF
%token MNODE
%token NIL
%token MNIL
%token CONS
%token MCONS
%token LBRACKET
%token RBRACKET
%token LTHAN
%token RTHAN
%token AND
%token MATCH
%token WITH
%token VDASH
%token SEMI
%token LBRACE
%token RBRACE
%token SLASH
%token IF
%token THEN
%token ELSE
%token TRUE
%token FALSE
%token QUOTE
%token TILDE
%token HAT
%token AT
%token DOT
%token MINUS
%token PC
%token ATAT
%token SHARP
/* %token TILDEPC */
%token DOLLAR
%token MOD

/* ̤ͥassociativity㤤⤤ء */

%right ARROW
%right IN
%right prec_if
/* %nonassoc TILDE */
/* %nonassoc FUN IF LTHAN LET */
%right prec_let
%right prec_fun
%nonassoc IDENT LPAREN BEGIN LBRACKET MATCH LBRACE AT DOT VDASH
%nonassoc TRUE FALSE STRING INT NIL MNIL
%left SEMI
%left SHARP
%right prec_node
%left prec_app
%right LEAF MLEAF
%right CONS MCONS
%right DOLLAR
%left ATAT MINUS EQUAL HAT RTHAN MOD
%right PC TILDEPC
%left prec_attrget
%nonassoc prec_label


/* ϵ */
%type <Syntax.t> file
%start file

%%

file:
  exp EOF
  { $1 }


exp_list:
| /* empty */
  { [] }
| exp
  { [$1] }
| exp SEMI exp_list
  { $1::$3 }

exp:
| LPAREN exp RPAREN
  { $2 }
| exp SHARP exp
  { BinOp(Semi, $1, $3) }
| BEGIN exp END
  { $2 }
| STRING
  { String $1 }
| INT
  { Int $1 }
| IDENT
  { Var $1 }
| TRUE { True }
| FALSE { False }
| exp MINUS exp { BinOp(Sub, $1, $3) }
| exp EQUAL exp { BinOp(Eq, $1, $3) }
| exp HAT exp { BinOp(Cat, $1, $3) }
| exp ATAT exp { BinOp(ListCat, $1, $3) }
| exp MOD exp { BinOp(Mod, $1, $3) }
| exp SHARP exp { BinOp(Semi, $1, $3) }
| FIX LPAREN IDENT COMMA IDENT COMMA exp RPAREN
  { Fix($3, $5, $7) }
| FUN IDENT ARROW exp
  { Lam($2, $4) }
| exp exp %prec prec_app
  { App($1, $2) }
| LET id EQUAL exp IN exp %prec prec_let
  { Let($2, $4, $6) }
| LET REC letrecs IN exp
  { Letrecs ($3, $5)}
| IF exp THEN exp ELSE exp %prec prec_if
  { If($2, $4, $6) }
| LTHAN IDENT RTHAN %prec prec_label
  { Label($2) }
| PC exp
  { Leaf $2 }
| exp AT exp DOLLAR exp %prec prec_node
  { Node($1, $3, $5) }
| LBRACE ident_eq_exp_list RBRACE
  { Attribute($2) }
| exp DOT IDENT %prec prec_attrget
  { AttrGet($1, $3) }
| NIL
  { Nil }
| exp CONS exp
  { Cons($1, $3) }
| LBRACKET exp_list RBRACKET
  { make_cons $2 }
| MATCH exp WITH matches
  { Match($2, $4) }

matches:
    one_match
    { [$1] }
|   one_match VDASH matches
    { $1::$3 }

one_match:
    pattern ARROW exp
    { $1, $3 }

pattern:
    LPAREN pattern RPAREN
    { $2 }
|    PC pattern
    { LeafPattern $2 }
|   pattern AT pattern DOLLAR pattern
    { 
      NodePattern($1, $3, $5)
    }
|   NIL
    { NilPattern }
|   pattern CONS pattern
    { ConsPattern($1, $3) }
|   id
    { IdentPattern $1 }
|   STRING
    { StringPattern $1 }
|   LTHAN IDENT RTHAN
    { LabelPattern $2 }

id:
    IDENT
    { $1 }

id_list:
    id
    { [$1] }
|   id id_list
    { $1::$2 }

letrec:
  IDENT EQUAL exp
  { $1, $3 }
  

letrecs:
    letrec
    { [$1] }
|  letrec AND letrecs
    { $1::$3 }

ident_eq_exp_list:
| /* empty */
  { [] }
| IDENT EQUAL exp
  { [$1, $3] }
| IDENT EQUAL exp SEMI ident_eq_exp_list
  { ($1, $3)::$5 }

/*
attr:
| LBRACE ident_eq_exp_list RBRACE
  { $2 }
*/
