
open Format
(* open Printf *)
open XmlLexer
open RuntimeType

exception EmptyStack

exception Eof_Exception

let lexbuf = ref (Lexing.from_string "")
let unread = ref None
let (stack : RuntimeType.token list ref) = ref []

let read_token () =
  match !unread with
      None -> token !lexbuf
    | Some x -> unread := None; x

let peek_token () =
  match !unread with
      None ->
	let x = token !lexbuf in
	  unread := Some x;
	  x
    | Some x -> x

let write_string l =
  print_string l;
  flush stdout

let pop () =
  match !stack with
      [] -> raise EmptyStack
    | hd::tl ->
	stack := tl;
	hd

let push x = stack := x::!stack

let rec write_attrs formatter = function
    [] -> ()
  | [k, v] -> fprintf formatter " %s=%s" (* " %s=%a" *) k (* write *) v
  | (k, v)::tl -> fprintf formatter " %s=%s%a" (* " %s=%a%a" *) k (* write *) v write_attrs tl

and write formatter = function
  | StartTag(x, attrs) -> fprintf formatter "<%s%a>" x write_attrs attrs
  | EndTag s -> fprintf formatter "</%s>" s
  | Pcdata s -> fprintf formatter "%s" s
  | _ -> failwith ""
  
let write = write std_formatter

let rec parse_tree unread =
  match unread with
      None -> parse_tree (Some (read_token ()))
    | Some l ->
	begin
	  match l with
	    StartTag(l, attr) -> Node(l, attr, parse_list ())
	  | Pcdata x -> Leaf x
	  | Eof -> raise Eof_Exception
	  | EndTag _ -> failwith "Illformed XML"
	end
and parse_list () =
  let l = read_token () in
    match l with
	EndTag l' -> []
      | _ -> 
	  try
	    let hd = parse_tree (Some l) in
	    let tl = parse_list () in
	      hd::tl
	  with
	      Eof_Exception -> []

let rec s2m () = parse_tree (Some (peek_token ()))
let rec sort f l = List.sort f l
let rec lessthan = compare

let rec print_attr chan formatter = function
    [] -> ()
  | [k, v] -> Printf.fprintf chan "%s=%s" k v
  | (k, v)::tl -> Printf.fprintf chan "%s=%s %a" k v (print_attr chan) tl
let rec print_list chan formatter = function
    [] -> ()
  | hd::tl ->
      Printf.fprintf chan "%a%a" (print_tree chan) hd (print_list chan) tl
and print_tree chan formatter = function
    Leaf x -> Printf.fprintf chan "%s" x
  | Node(l, attr, tl) ->
      match attr with [] -> Printf.fprintf chan "<%s>%a</%s>" l (print_list chan) tl l
	| _ -> Printf.fprintf chan "<%s %a>%a</%s>" l (print_attr chan) attr (print_list chan) tl l

let print_list  = print_list stdout stdout

let rec copy_list () =
  let l = peek_token () in
    match l with
	(StartTag(l', at)) as tag -> (* write tag; *) copy_tree (); copy_list ()
      | EndTag l -> (* assert(pop() = l); *)ignore(read_token ())
      | Pcdata s -> read_token (); write_string s; copy_list ()
and copy_tree () =
  let l = peek_token () in
    match l with
	(StartTag(l', at)) as tag -> read_token (); write tag; copy_list (); read_token (); write (EndTag l')
      | Pcdata x -> read_token (); write_string x
      | EndTag _ -> () (* ignore(read_token ()) *)

let rec skip_list () =
  let l = peek_token () in
    match l with
	(StartTag(l', at)) as tag -> (* write tag; *) skip_tree (); skip_list ()
      | EndTag l -> () (* assert(pop() = l); *)(* ignore(read_token ()) *)
      | Pcdata s -> read_token (); skip_list ()
and skip_tree () =
  let l = peek_token () in
    match l with
	(StartTag(l', at)) as tag -> read_token (); skip_list (); (ignore(read_token ()))
      | Pcdata x -> ignore(read_token ())
      | EndTag _ -> () (* ignore(read_token ()) *)

let pc_of_int n = write (Pcdata (string_of_int n))
let int_of_pc = function Pcdata s -> int_of_string s
