open Typeinf

let subst_use u_source u_target c =
  match c with
      Ge_use(u1, u2) ->
	let u1' = if u1 = Alpha.UVar u_target then u_source else u1 in
	let u2' = if u2 = Alpha.UVar u_target then u_source else u2 in
	  Ge_use(u1', u2')
    | Eq_use(u1, u2) ->
	let u1' = if u1 = Alpha.UVar u_target then u_source else u1 in
	let u2' = if u2 = Alpha.UVar u_target then u_source else u2 in
	  Eq_use(u1', u2')

(* debug *)
let rec print_ans ans =
  match ans with
      [] -> ()
    | (uid, u)::tl ->
	print_string "(";
	print_int uid;
	print_string "=";
	print_usage u;
	print_string ") ";
	print_ans tl

let rec subst_ans ans substed_list =
  match substed_list with
      (uvar1, uvar2)::tl ->
	if (List.mem uvar1 (List.map fst ans)) then
	  subst_ans ((uvar2, (List.assoc uvar1 ans))::ans) tl
	else subst_ans ans tl
    | [] -> ans


let rec solver_iter tobe_cs done_cs ans substed substed_list =
  match tobe_cs with
	[] ->
	  if substed then
	    solver_iter done_cs [] ans false substed_list
	  else
	    subst_ans ans substed_list
    | (Ge_use(u1, u2))::tl ->
	begin
	  match u1, u2 with
	    | Alpha.UOmega, u1 ->
		solver_iter tl done_cs ans substed substed_list
	    | u1, u2 when u1 = u2 ->
		solver_iter tl done_cs ans substed substed_list
	    | Alpha.UVar u, Alpha.UOmega ->
		let tl' = List.map (subst_use Alpha.UOmega u) tl in
		let done_cs' = List.map (subst_use Alpha.UOmega u) done_cs in
		  solver_iter tl' done_cs' ((u,Alpha.UOmega)::ans) true substed_list
	    | Alpha.UVar uvar1, Alpha.UVar uvar2 ->
		solver_iter tl (Ge_use(Alpha.UVar uvar1, Alpha.UVar uvar2)::done_cs) ans substed substed_list
		  (* UOneがらみのマッチング。 *)
	    | u1, Alpha.UOne ->
		solver_iter tl done_cs ans substed substed_list
	    | Alpha.UOne, Alpha.UVar u ->
		let tl' = List.map (subst_use Alpha.UOne u) tl in
		let done_cs' = List.map (subst_use Alpha.UOne u) done_cs in
		  solver_iter tl' done_cs' ((u,Alpha.UOne)::ans) true substed_list
	    | _ -> failwith "solve(c)"
	end  
    | (Eq_use(u1, u2))::tl ->
	begin
	  match u1, u2 with
	      use1, use2 when use1 = use2 -> solver_iter tl done_cs ans substed substed_list 
	    | Alpha.UVar uvar1, Alpha.UVar uvar2 -> 
		let tl' = List.map (subst_use (Alpha.UVar uvar1) uvar2) tl in
		let done_cs' = List.map (subst_use (Alpha.UVar uvar1) uvar2) done_cs in
		  solver_iter tl' done_cs' ans true ((uvar1, uvar2)::substed_list)
	    | Alpha.UVar uvar1, u2 ->
		let tl' = List.map (subst_use u2 uvar1) tl in
		let done_cs' = List.map (subst_use u2 uvar1) done_cs in
		  solver_iter tl' done_cs' ((uvar1, u2)::ans) true substed_list
	    | u1, Alpha.UVar uvar2 ->
		let tl' = List.map (subst_use u1 uvar2) tl in
		let done_cs' = List.map (subst_use u1 uvar2) done_cs in
		  solver_iter tl' done_cs' ((uvar2, u1)::ans) true substed_list
	    | _ -> failwith "can't solve constr"
	end
	  
(* solve(c)*)
let rec solver (constrs : Typeinf.constr list)=
  solver_iter constrs [] [] false []

let find_use u ans =
  try
    List.assoc u ans
  with
    | _ -> Alpha.UOne

let rec translate t constrs =
  let ans = solver constrs in
    (trans_inner ans t)


and trans_inner ans t =
  let t, typ = t in
  let pre_t =
    match t with
	String s -> t
      | Int n -> t
      | True | False -> t
      (* | Pcdata pc -> t *)
      | Label l -> t
      | Attribute a -> t
      | AttrGet(a, x) -> 
	  AttrGet((trans_inner ans a), x)
      | True -> t
      | False -> t
      | BinOp (binop, t1, t2) ->
	  BinOp(binop, trans_inner ans t1, trans_inner ans t2)
      | Var x -> t
      | Fix(f, x, t') ->
	  Fix(f, x, trans_inner ans t')
      | App(t1, t2) ->
	  App(trans_inner ans t1, trans_inner ans t2)
      | Let(x, t1, t2) ->
	  Let(x, trans_inner ans t1, trans_inner ans t2)
      | Letrecs(fts, t) ->
	  let rec trans_inner_match ans fts =
	    (match fts with
		 (fn, tn)::tl ->
		   (fn, (trans_inner ans tn))::(trans_inner_match ans tl)
	       | [] -> [])
	  in 
	  let fts' = trans_inner_match ans fts in
	  let t' = trans_inner ans t in
	    Letrecs(fts', t')
      | If(t0, t1, t2) ->
	  If(trans_inner ans t0, trans_inner ans t1, trans_inner ans t2)
      | Leaf(u, t) ->
	  Leaf(translate_use u ans, trans_inner ans t)
      | Node(u, t1, t2, t3) ->
	  Node(translate_use u ans, trans_inner ans t1, trans_inner ans t2, trans_inner ans t3)
      | Nil(u) -> Nil(translate_use u ans)
      | Cons(u, t1, t2) -> 
	  Cons(translate_use u ans, trans_inner ans t1, trans_inner ans t2)
(**)
      | Match(typ, id, match_list) -> 
	  let rec trans_inner_match ans (p, t) =
	    (p, (trans_inner ans t))
	  in
	  (
	  match typ with 
	      Alpha.TTree u ->
		Match(Alpha.TTree (translate_use u ans), id, (List.map (trans_inner_match ans) match_list))
	    | Alpha.TTreelist u ->
		Match(Alpha.TTreelist (translate_use u ans), id, (List.map (trans_inner_match ans) match_list))
	    | _ -> Match(typ, id, (List.map (trans_inner_match ans) match_list))
	  ) 
      | Coerce(env1, env2, t) ->
	  let env1' = List.map (fun (x,t) -> (x,translate_type t ans)) env1 in
	  let env2' = List.map (fun (x,t) -> (x,translate_type t ans)) env2 in
	    translate_coerce env1' env2' (trans_inner ans t)
      | LetS2M_t(_, _) -> failwith "translate: LetS2M_t is not allowed to occur here"
      | LetS2M_tl(_, _) -> failwith "translate: LetS2M_tl is not allowed to occur here"

  in
    pre_t, (translate_type typ ans) 
(*
  | M2S _ -> failwith "translate: M2S is not allowed to occur here"
  | Copy _ -> failwith "translate: Copy is not allowed to occur here"
*)
and translate_type typ ans =
  match typ with
      Alpha.TString -> Alpha.TString
    | Alpha.TInt -> Alpha.TInt
    | Alpha.TBool -> Alpha.TBool
    | Alpha.TLabel -> Alpha.TLabel
    | Alpha.TAttribute -> Alpha.TAttribute
    | Alpha.TFun(t1, t2) ->
	Alpha.TFun(translate_type t1 ans, translate_type t2 ans)
    | Alpha.TTree u -> Alpha.TTree (translate_use u ans)
    | Alpha.TTreelist u -> Alpha.TTreelist (translate_use u ans)

and translate_use u ans =
  match u with
      Alpha.UOne | Alpha.UOmega -> u
    | Alpha.UVar u -> find_use u ans
and translate_coerce env1 env2 t =    (*coerce env env' t*)
  let pre_t, typ = t in
    match env1, env2 with
	[], [] -> pre_t
      | (x1,typ1)::tl1, (x2,typ2)::tl2 when x1 = x2 ->
	  begin
	    match typ1, typ2 with
		Alpha.TString,  Alpha.TString
	      | Alpha.TInt, Alpha.TInt
	      | Alpha.TBool, Alpha.TBool
	      | Alpha.TLabel, Alpha.TLabel | Alpha.TAttribute, Alpha.TAttribute -> translate_coerce tl1 tl2 t
	      | Alpha.TFun(t1d, t1cod), Alpha.TFun(t2d, t2cod) when t1d = t2d && t1cod = t2cod ->
		  translate_coerce tl1 tl2 t
	      | Alpha.TTree(Alpha.UOmega), Alpha.TTree(Alpha.UOmega)
	      | Alpha.TTree(Alpha.UOne), Alpha.TTree(Alpha.UOne) ->
		  translate_coerce tl1 tl2 t
	      | Alpha.TTree(Alpha.UOne), Alpha.TTree(Alpha.UOmega) ->
		  LetS2M_t(x1, (translate_coerce tl1 tl2 t, typ))
	      | Alpha.TTreelist(Alpha.UOmega), Alpha.TTreelist(Alpha.UOmega)
	      | Alpha.TTreelist(Alpha.UOne), Alpha.TTreelist(Alpha.UOne) ->
		  translate_coerce tl1 tl2 t
	      | Alpha.TTreelist(Alpha.UOne), Alpha.TTreelist(Alpha.UOmega) ->
		  LetS2M_tl(x1, (translate_coerce tl1 tl2 t, typ))
(*
	      | _ -> Typeinf.print_typ typ1;
		  print_newline();
		  Typeinf.print_typ typ2;
		  print_newline();
		  Typeinf.print_term 0 t;
		  print_newline();
		  failwith "translate_coerce: CANNOT OCCUR!!"
*)
	  end
      | _ -> failwith "translate_coerce: CANNOT OCCUR!"
