type ident = string
type pcdata = string
type label = string

type binop = Eq | Cat | Sub | ListCat | Semi | Mod

type t =
    String of string
  | Int of int
  | True | False
  (* | Pcdata of pcdata *)
  | Attribute of attribute
  | Label of label
  | BinOp of binop * t * t
  | Var of ident
  | Fix of ident * ident * t (* Fix of ident list * (t list)*)
  | App of t * t
  | If of t * t * t
  | Leaf of t
  | Node of t * t * t
  | Nil
  | Cons of t * t
  | Match of t * one_match list
  | Let of ident * t * t
  | Lam of ident * t
  | Letrecs of (ident * t) list * t
  | AttrGet of t * ident
and attribute = (string * t) list
and one_match = (pattern * t)
and pattern =
    LeafPattern of (*pcdata*) pattern
  | NodePattern of (*label*) pattern * (*attribute*)pattern * (*treelist*)pattern
  | NilPattern
  | ConsPattern of (*tree*)pattern * (*treelist*)pattern
  | IdentPattern of ident
  | StringPattern of string
  | LabelPattern of label

let rec term_to_string t =
      match t with
	  String s -> ("\""^s^"\"")
      (* | Pcdata pc -> pc *)
	| Attribute a -> "Attr"
	| Int n -> string_of_int n
	| Label l -> l
	| Var id -> id
	| Fix (id1, id2, t) ->
	    "fix("^id1^", "^"id2"^", \n"^(term_to_string t)^")"
	| App (t1, t2) ->
	    "("^(term_to_string t1)^" "^(term_to_string t2)^")";
	| Leaf t0 -> 
	    "Leaf "^(term_to_string t0)
	| Node (t1, t2, t3) ->
	    "Node("^(term_to_string t1)^", "^(term_to_string t2)^", "^(term_to_string t3)^")"
	| Nil -> "Nil"
	| Cons (t1, t2) ->
	    "Cons("^(term_to_string t1)^", "^(term_to_string t2)^")"
	| Match (t, match_list) ->
	    "Match("^(term_to_string t)^", "^(matchs_to_string match_list)^")"
	| _ -> "other"
and pattern_to_string p =
  match p with
      LeafPattern _ -> "leaf"
    | NodePattern _ -> "node"
    | NilPattern -> "nil"
    | ConsPattern _ -> "cons"
    | StringPattern _ -> "str"
    | IdentPattern id -> "id_"^id
    | LabelPattern _ -> "label"
and matchs_to_string ms =
  match ms with
      [] -> "[]"
    | (p, t)::tl ->
	"("^(pattern_to_string p)^", "^(term_to_string t)^")::"^(matchs_to_string tl)





(*
let rec print_t t =
      match t with
	  String s -> print_string "Str"
  (* | Pcdata pc -> print_string "Pc" *)
	| Attribute a -> print_string "Attr"
	| Label l -> print_string "Lbl"
	| Var id -> print_string id
	| Fix (id1, id2, t) ->
	    print_string "fix(";
	    print_string id1;
	    print_string ", ";
	    print_string id2;
	    print_string ", \n";
	    print_t t;
	    print_string ")"
	| App (t1, t2) ->
	    print_string "App(";
	    print_t t1;
	    print_string ", ";
	    print_t t2;
	    print_string ")";
	| Leaf t0 -> 
	    print_string "Leaf ";
	    print_t t0
	| Node (t1, t2, t3) ->
	    print_string "Node(";
	    print_t t1;
	    print_string ", ";
	    print_t t2;
	    print_string ", ";
	    print_t t3;
    	    print_string ")";
	| Nil -> print_string "Nil"
	| Cons (t1, t2) ->
	    print_string "Cons(";
	    print_t t1;
	    print_string ", ";
	    print_t t2;
    	    print_string ")";	    
	| Match (t, match_list) ->
	    print_string "Match(";
	    print_t t;
	    print_string ", ";
	    print_matchs match_list;
	    print_string ")"
	| _ -> print_string "other"
and print_pattern p =
  match p with
      LeafPattern _ ->
	print_string "leaf"
    | NodePattern _ ->
	print_string "node"
    | NilPattern ->
	print_string "nil"
    | ConsPattern _ ->
	print_string "cons"
    | StringPattern _ ->
	print_string "str"
    | IdentPattern id ->
	print_string "id_"; print_string id;
    | LabelPattern _ ->
	print_string "label"
and print_matchs ms =
  match ms with
      [] -> print_string "[]"
    | (p, t)::tl ->
	print_string "(";
	print_pattern p;
	print_string ", ";
	print_t t;
	print_string ")::";
	print_matchs tl
*)
