open Typeinf


type m2s =
    M2S of treetype
and treetype = T | TL


let trans_typ typ =
  match typ with
      Alpha.TTree Alpha.UOmega -> Alpha.TTree Alpha.UPlus
    | Alpha.TTreelist Alpha.UOmega -> Alpha.TTreelist Alpha.UPlus

let insert_m2s t_tl (term, typ) =
  if t_tl = T then
    M2S_t (term, typ), trans_typ typ
  else
    M2S_tl (term, typ), trans_typ typ


let rec trans_m2s (term, typ) =
  match term with
      M2S_t(t, _) ->
	if (enable t) then
	  trans_inner t T, typ
	else 
	  term, typ
    | M2S_tl(t, _) ->
	if (enable t) then
	  trans_inner t TL, typ
	else 
	  term, typ
    | Fix(f, x, t) ->
	Fix(f, x, (trans_m2s t)), typ
    | App(t1, t2) ->
	App((trans_m2s t1), (trans_m2s t2)), typ
    | Let(x, t1, t2) ->
	Let(x, (trans_m2s t1), (trans_m2s t2)), typ
    | Letrecs(fts, t) ->
	let rec trans_fts fts =
	  (match fts with
	       (f, t)::tl -> 
		 (f, trans_m2s t)::(trans_fts tl)
	     | [] -> [])
	in
	Letrecs(trans_fts fts, (trans_m2s t)), typ
    | Match(typ, y, pts) ->
	let rec trans_pts pts =
	  (match pts with
	       (p, t)::tl -> (p, trans_m2s t)::(trans_pts tl)
	     | [] -> [])
	in
	  Match(typ, y, trans_pts pts), typ
    | LetS2M_t(x, t) ->
	LetS2M_t(x, trans_m2s t), typ
    | LetS2M_tl(x, t) ->
	LetS2M_tl(x, trans_m2s t), typ
    | _ -> (term, typ)
and trans_inner t t_tl=
  match t with
      Leaf(Alpha.UOmega, t1) ->
	Leaf(Alpha.UPlus, t1)
    | Node(Alpha.UOmega, t1, t2, t3) ->
	Node(Alpha.UPlus, t1, t2, (trans_m2s (M2S_tl t3, (snd t3))))
    | Cons(Alpha.UOmega, t1, t2) ->
	Cons(Alpha.UPlus, (trans_m2s (M2S_t t1, (snd t1))), (trans_m2s (M2S_tl t2, (snd t2))))
    | Nil(Alpha.UOmega) -> Nil(Alpha.UPlus)
    | Match(typ, y, pts) ->
	let rec f pts =
	  (match pts with
	       (p, t)::tl ->
		 let t' = (insert_m2s t_tl t) in
		   (p, trans_m2s t')::(f tl)
	     | [] -> [])
	in 
	  Match(typ, y, f pts)
    | Let(x, t1, t2) ->
	let t2' = insert_m2s t_tl t2 in
	  Let(x, trans_m2s t1, trans_m2s t2')
    | Letrecs(fts, t) ->
	let t' = insert_m2s t_tl t in
	  Letrecs(fts, trans_m2s t')
    | LetS2M_t(x, t) ->
	let t' = insert_m2s t_tl t in
	  LetS2M_t(x, trans_m2s t')
    | LetS2M_tl(x, t) ->
	let t' = insert_m2s t_tl t in
	  LetS2M_tl(x, trans_m2s t')
and enable t =
  match t with
      Leaf _ | Node _ | Cons _ | Nil _ | Match _ | Let _ | Letrecs _ | LetS2M_t _ | LetS2M_tl _ -> true
    | _ -> false
and typ_of pts =
  match pts with
      (p, (t, typ))::tl -> typ
and trans_m2st_pt (p, t) =
  p, (trans_m2s (M2S_t t, (snd t)))
and trans_m2stl_pt (p, t) =
  p, (trans_m2s (M2S_tl t, (snd t)))









let rec trans_copy t =
  let rec inner t0 =
    match t0 with
	LetS2M_t(x, (M2S_t (Var y, _), _)), typ when x=y -> (Copy x), typ
      | LetS2M_tl(x, (M2S_tl (Var y, _), _)), typ when x=y -> (Copy x), typ
      | LetS2M_t(x, t), typ -> LetS2M_t(x, trans_copy t), typ
      | LetS2M_tl(x, t), typ -> LetS2M_tl(x, trans_copy t), typ
      | _ -> t0
  in
    match t with
	Fix(f, x, t), typ ->
	  Fix(f, x, trans_copy t), typ
      | Let(x, t1, t2), typ ->
	  Let(x, trans_copy t1, trans_copy t2), typ
      | Letrecs(fts, t), typ ->
	  let a (f, t) = (f, trans_copy t) in
	    Letrecs(List.map a fts, trans_copy t), typ
      | Match(typ_x, x, pts), typ ->
	  let a (p,t) = (p, (trans_copy t)) in
	    Match(typ_x, x, (List.map a pts)), typ
      | Node(u, l, a, tl), typ ->
	  Node(u, l, a, (trans_copy tl)), typ
      | _ -> inner t




let rec subtract id ids =
  List.filter (fun x -> x <> id) ids

let rec subtracts ids1 ids2 =
  match ids1 with
      [] -> ids2
    | hd::tl -> subtracts tl (subtract hd ids2)

let rec fv (t, typ) =
  match t with
    BinOp (binop, t1, t2) -> (fv t1)@(fv t2) 
  | Var id -> [id]
  | Fix (f, (x, _), t) ->
      subtract x (subtract f (fv t))
  | App (t1, t2) -> (fv t1)@(fv t2)
  | If (t0, t1, t2) -> (fv t0)@(fv t1)@(fv t2)
  | Let ((x,_), t1, t2) ->
      subtract x ((fv t1)@(fv t2))
  | Letrecs(fts, t) ->
      let fs = List.map fst fts in
      let ts = List.map snd fts in
      let rec fv_ts ts =
	match ts with
	    [] -> []
	  | hd::tl -> (fv hd)@(fv_ts tl)
      in 
	subtracts fs ((fv_ts ts)@(fv t))
  | Leaf (u, t) -> fv t
  | Node (u, t_label, t_attr, t_treelist)-> 
      (fv t_label)@(fv t_attr)@(fv t_treelist)
  | Nil u -> []
  | Cons (u, t_hd, t_tl) -> (fv t_hd)@(fv t_tl)
  | Match (typ, id, matchs) ->
      id::(fv_matchs matchs)
  | LetS2M_t (x, t) -> fv t
  | LetS2M_tl (x, t) -> fv t
  | M2S_t t -> fv t
  | M2S_tl t -> fv t
  (* | Pcdata _ -> [] *)
  | String _ -> []
  | Int _ -> []
  | True | False -> []
  (* | Pcdata _ -> [] *)
  | Attribute _ -> []
  | Label _ -> []
  | Copy _ -> []
and fv_matchs matchs =
  match matchs with
      [] -> []
    | (p, t)::tl ->
	(subtracts (fv_pattern p) (fv t))@(fv_matchs tl)
and fv_pattern p =
  match p with
      Alpha.LeafPattern p -> fv_pattern p
    | Alpha.NodePattern (p1, p2, p3) ->
	(fv_pattern p1)@(fv_pattern p2)@(fv_pattern p3)
    | Alpha.NilPattern -> []
    | Alpha.ConsPattern (p1, p2) -> (fv_pattern p1)@(fv_pattern p2)
    | Alpha.IdentPattern id -> [id]
    | Alpha.StringPattern str -> []
    | Alpha.LabelPattern l -> []


let rec trans_skip (t, typ) =
  let output_term = 
    match t with
	LetS2M_t(x, t') ->
	  let fvs = fv t' in
	    if (List.mem x fvs) then
	      LetS2M_t(x, trans_skip t')
	    else
	      Skip(x, trans_skip t')    
      | LetS2M_tl(x, t') ->
	  let fvs = fv t' in
	    if (List.mem x fvs) then
	      LetS2M_tl(x, trans_skip t')
	    else
	      Skip(x, trans_skip t')
      | BinOp(binop, t1, t2) -> BinOp(binop, trans_skip t1, trans_skip t2)
      | Fix(f, x, t) -> Fix(f, x, trans_skip t)
      | App(t1, t2) -> App(trans_skip t1, trans_skip t2)
      | Let(x, t1, t2) -> Let(x, trans_skip t1, trans_skip t2)
      | Letrecs(fts, t) ->
	  let rec a fts =
	    (match fts with
		 (fn, tn)::tl -> (fn, trans_skip tn)::(a tl)
	       | [] -> [])
	in
	    Letrecs(a fts, trans_skip t)
      | If(t0, t1, t2) ->
	  If(trans_skip t0, trans_skip t1, trans_skip t2)
      | Leaf (u, t) ->
	  Leaf(u, trans_skip t)
      | Node(u, t1, t2, t3) ->
	  Node(u, trans_skip t1, trans_skip t2, trans_skip t3)
      | Cons(u, t1, t2) ->
	  Cons(u, trans_skip t1, trans_skip t2)
      | Match(typ_match, y, pts) ->
	  let rec a pts =
	    (match pts with
		 (p, t)::tl -> (p, trans_skip t)::(a tl)
	       | [] -> [])
	in 
	    Match(typ_match, y, a pts)
      | M2S_t t ->
	  M2S_t (trans_skip t)
      | M2S_tl t ->
	  M2S_tl (trans_skip t)
      | _-> t
  in 
    output_term, typ
